<?php

declare(strict_types=1);

/**
 * @author Amasty Team
 * @copyright Copyright (c) Amasty (https://www.amasty.com)
 * @package Rules Tiered Discount by Amasty
 */

namespace Amasty\RulesTieredDiscount\Model\Rule\Action\Discount;

use Amasty\Rules\Model\Rule\Action\Discount\AbstractRule;
use Amasty\RulesTieredDiscount\Model\TieredDiscount;
use Magento\Catalog\Model\Product\Type;
use Magento\Framework\Phrase;
use Magento\Quote\Model\Quote\Item\AbstractItem;
use Magento\SalesRule\Model\Rule;
use Magento\SalesRule\Model\Rule\Action\Discount\Data;

class TieredDiscountPercent extends AbstractRule
{
    public const ACTION_NAME = 'tiered_discount_percent';

    /**
     * @param Rule $rule
     * @param AbstractItem $item
     * @param float $qty
     * @return Data
     */
    public function calculate($rule, $item, $qty): Data
    {
        $this->beforeCalculate($rule);
        $discountData = $this->calculateDiscount($rule, $item);
        $this->afterCalculate($discountData, $rule, $item);

        return $discountData;
    }

    public function getActionLabel(): Phrase
    {
        return __('Percent Discount: Buy X (Buy 1 for 10% off,  Buy 2 for 20% off, etc)');
    }

    /**
     * @param Rule $rule
     * @param AbstractItem $item
     * @return bool
     */
    public function skip($rule, $item): bool
    {
        $parent = $item->getParentItem();

        if ($parent && $parent->getProduct()->getTypeId() === Type::TYPE_BUNDLE) {
            return true;
        }

        return parent::skip($rule, $item);
    }

    private function calculateDiscount(Rule $rule, AbstractItem $item): Data
    {
        $amrule = $rule->getData(self::AMASTY_RULE);
        $tieredDiscounts = $this->prepareData($rule->getExtensionAttributes()->getAmrulesTieredDiscount());
        $allItems = $this->getSortedItems($item->getAddress(), $rule, $amrule->getApplyDiscountTo());
        $store = $item->getQuote()->getStore();
        $minQty = $tieredDiscounts[0]['quantity'];
        $maxQty = end($tieredDiscounts)['quantity'];
        $discountData = $this->discountFactory->create();
        [$baseItemPrice, $baseItemOriginalPrice] = $this->getItemPrices($item);
        $maxRuleAppliedQty = (int)$rule->getDiscountQty();
        $ruleAppliedQty = 0;

        $allItemsQty = count($allItems);
        while ($allItemsQty >= $minQty) {
            while ($allItemsQty < $maxQty) {
                array_pop($tieredDiscounts);
                $maxQty = end($tieredDiscounts)['quantity'];
            }
            $qtyToDiscount = (int)(floor($allItemsQty / $maxQty) * $maxQty);

            if ($maxRuleAppliedQty && ($qtyToDiscount/$maxQty + $ruleAppliedQty) > $maxRuleAppliedQty) {
                $qtyToDiscount = (int)($maxQty * ($maxRuleAppliedQty - $ruleAppliedQty));
            }

            $itemsForDiscount = array_slice($allItems, 0, $qtyToDiscount);
            $percentage = min(100, end($tieredDiscounts)['discount']) / 100;

            foreach ($itemsForDiscount as $itemForDiscount) {
                if ($item->getAmrulesId() === $itemForDiscount->getAmrulesId()
                    || $this->isValidChildOfBundle($item, [$itemForDiscount->getAmrulesId()])
                ) {
                    $baseDiscount = $baseItemPrice * $percentage;
                    $itemDiscount = $this->priceCurrency->convert($baseDiscount, $store);
                    $baseOriginalDiscount = $baseItemOriginalPrice * $percentage;
                    $originalDiscount = $this->priceCurrency->convert($baseOriginalDiscount, $store);

                    $itemDiscount = $this->itemPrice->resolveFinalPriceRevert($itemDiscount, $item);
                    $baseDiscount = $this->itemPrice->resolveBaseFinalPriceRevert($baseDiscount, $item);

                    $discountData->setAmount($discountData->getAmount() + $itemDiscount);
                    $discountData->setBaseAmount($baseDiscount + $discountData->getBaseAmount());
                    $discountData->setOriginalAmount($originalDiscount + $discountData->getOriginalAmount());
                    $discountData->setBaseOriginalAmount(
                        $baseOriginalDiscount + $discountData->getBaseOriginalAmount()
                    );
                }
            }
            $ruleAppliedQty += $qtyToDiscount/$maxQty;
            if ($maxRuleAppliedQty && ($ruleAppliedQty >= $maxRuleAppliedQty)) {
                return $discountData;
            }

            $allItems = array_slice($allItems, $qtyToDiscount);
            $allItemsQty = count($allItems);
        }

        return $discountData;
    }

    /**
     * @param TieredDiscount[] $tieredDiscountsEntities
     * @return array [['quantity' => {qty}, 'discount' => {discount}], ...]
     */
    private function prepareData(array $tieredDiscountsEntities): array
    {
        $tieredDiscountsData = [];
        foreach ($tieredDiscountsEntities as $entity) {
            $tieredDiscountsData[] = [
                'quantity' => $entity->getQuantity(),
                'discount' => $entity->getDiscount()
            ];
        }

        return $tieredDiscountsData;
    }

    /**
     * @param AbstractItem $item
     * @return float[]
     */
    private function getItemPrices(AbstractItem $item): array
    {
        $baseItemPrice = $this->itemPrice->getItemBasePrice($item);
        $baseItemOriginalPrice = $this->itemPrice->getItemBaseOriginalPrice($item);

        $parent = $item->getParentItem();
        if ($parent && $parent->getProduct()->getTypeId() === Type::TYPE_BUNDLE) {
            $baseItemPrice *= $item->getQty();
            $baseItemOriginalPrice *= $item->getQty();
        }

        return [$baseItemPrice, $baseItemOriginalPrice];
    }
}
